# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import json
import os
import pathlib
import tempfile


def package_name_from_interface_file_path(path):
    """
    Derive ROS package name from a ROS interface definition file path.

    This function assumes ROS interface definition files follow the typical
    ``rosidl`` install space layout i.e. 'package_name/subfolder/interface.idl'.
    """
    return pathlib.Path(os.path.abspath(path)).parents[1].name


def dependencies_from_include_paths(include_paths):
    """
    Collect dependencies' ROS interface definition files from include paths.

    Interface definition file paths from dependencies are absolute paths
    prefixed by the name of package they belong to followed by a colon ':'.
    """
    return list({
        f'{package_name_from_interface_file_path(path)}:{path}'
        for include_path in include_paths
        for path in pathlib.Path(
            os.path.abspath(include_path)
        ).glob('**/*.idl')
    })


def interface_path_as_tuple(path):
    """
    Express interface definition file path as an (absolute prefix, relative path) tuple.

    An interface definition file path is a relative path, optionally prefixed
    by a path against which to resolve the former followed by a colon ':'.
    Thus, this function applies following logic:
    - If a given path follows this pattern, it is split at the colon ':'
    - If a given path is prefixed by a relative path, it is resolved
      relative to the current working directory.
    - If a given path has no prefix, the current working directory is
      used as prefix.
    """
    path_as_string = str(path)
    if ':' not in path_as_string:
        prefix = pathlib.Path.cwd()
    else:
        prefix, _, path = path_as_string.rpartition(':')
        prefix = pathlib.Path(os.path.abspath(prefix))
    path = pathlib.Path(path)
    if path.is_absolute():
        raise ValueError('Interface definition file path '
                         f"'{path}' cannot be absolute")
    return prefix, path


def idl_tuples_from_interface_files(interface_files):
    """
    Express ROS interface definition file paths as IDL tuples.

    An IDL tuple is a relative path prefixed by an absolute path against
    which to resolve it followed by a colon ':'. This function then applies
    the same logic as `interface_path_as_tuple`.
    """
    idl_tuples = []
    for path in interface_files:
        prefix, path = interface_path_as_tuple(path)
        idl_tuples.append(f'{prefix}:{path.as_posix()}')
    return idl_tuples


@contextlib.contextmanager
def legacy_generator_arguments_file(
    *,
    package_name,
    interface_files,
    include_paths,
    templates_path,
    output_path
):
    """
    Generate a temporary rosidl generator arguments file.

    :param package_name: Name of the ROS package for which to generate code
    :param interface_files: Relative paths to ROS interface definition files,
      optionally prefixed by another absolute or relative path followed by
      a colon ':'. The former relative paths will be used as a prototype to
      namespace generated code (if applicable).
    :param include_paths: Paths where ROS package dependencies' interface
      definition files may be found
    :param templates_path: Path to the templates directory for the
      generator script this arguments are for
    :param output_path: Path to the output directory for generated code
    """
    idl_tuples = idl_tuples_from_interface_files(interface_files)
    interface_dependencies = dependencies_from_include_paths(include_paths)
    output_path = os.path.abspath(output_path)
    templates_path = os.path.abspath(templates_path)
    # NOTE(hidmic): named temporary files cannot be opened twice on Windows,
    # so close it and manually remove it when leaving the context
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp:
        tmp.write(json.dumps({
            'package_name': package_name,
            'output_dir': output_path,
            'template_dir': templates_path,
            'idl_tuples': idl_tuples,
            'ros_interface_dependencies': interface_dependencies,
            # TODO(hidmic): re-enable output file caching
            'target_dependencies': []
        }))
    path_to_file = os.path.abspath(tmp.name)
    try:
        yield path_to_file
    finally:
        try:
            os.unlink(path_to_file)
        except FileNotFoundError:
            pass


def generate_visibility_control_file(
    *,
    package_name,
    template_path,
    output_path
):
    """
    Generate a visibility control file from a template.

    :param package_name: Name of the ROS package for which
      to generate the file.
    :param template_path: Path to template visibility control file.
      May contain @PROJECT_NAME@ and @PROJECT_NAME_UPPER@ placeholders,
      to be substituted by the package name, accordingly.
    :param output_path: Path to visibility control file after interpolation.
    """
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    with open(template_path, 'r') as fd:
        content = fd.read()

    content = content.replace('@PROJECT_NAME@', package_name)
    content = content.replace('@PROJECT_NAME_UPPER@', package_name.upper())

    with open(output_path, 'w') as fd:
        fd.write(content)



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rosidl_cli.extensions import Extension
from rosidl_cli.extensions import load_extensions


class GenerateCommandExtension(Extension):
    """
    The extension point for source code generation.

    The following methods must be defined:
    * `generate`
    """

    def generate(
        self,
        package_name,
        interface_files,
        include_paths,
        output_path
    ):
        """
        Generate source code.

        Paths to interface definition files are relative paths optionally
        prefixed by an absolute path followed by a colon ':', in which case
        path resolution is to be performed against that absolute path.

        :param package_name: name of the package to generate source code for
        :param interface_files: list of paths to interface definition files
        :param include_paths: list of paths to include dependency interface
          definition files from.
        :param output_path: path to directory to hold generated source code files
        :returns: list of paths to generated source files
        """
        raise NotImplementedError()


def load_type_extensions(**kwargs):
    """Load extensions for type representation source code generation."""
    return load_extensions('rosidl_cli.command.generate.type_extensions', **kwargs)


def load_typesupport_extensions(**kwargs):
    """Load extensions for type support source code generation."""
    return load_extensions('rosidl_cli.command.generate.typesupport_extensions', **kwargs)



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Command:
    """
    The extension point for 'command' extensions.

    The following methods must be defined:
    * `main`
    * `add_arguments`
    """

    def add_arguments(self, parser):
        pass

    def main(self, *, parser, args):
        raise NotImplementedError()



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata


logger = logging.getLogger(__name__)


def get_entry_points(group_name, *, specs=None, strict=False):
    """
    Get entry points from a specific group.

    :param str group_name: the name of the entry point group
    :param list specs: an optional collection of entry point names to retrieve
    :param bool strict: whether to raise or warn on error
    :returns: mapping from entry point names to ``EntryPoint`` instances
    :rtype: dict
    """
    if specs is not None:
        specs = set(specs)
    entry_points_impl = importlib_metadata.entry_points()
    if hasattr(entry_points_impl, 'select'):
        groups = entry_points_impl.select(group=group_name)
    else:
        groups = entry_points_impl.get(group_name, [])
    entry_points = {}
    for entry_point in groups:
        name = entry_point.name
        if specs and name not in specs:
            continue
        if name in entry_points:
            msg = (f"Found duplicate entry point '{name}': "
                   f'got {entry_point} and {entry_points[name]}')
            if strict:
                raise RuntimeError(msg)
            logger.warning(msg)
            continue
        entry_points[name] = entry_point
    if specs:
        pending = specs - set(entry_points)
        if pending:
            msg = 'Some specs could not be met: '
            msg += ', '.join(map(str, pending))
            if strict:
                raise RuntimeError(msg)
            logger.warning(msg)
    return entry_points


def load_entry_points(group_name, *, strict=False, **kwargs):
    """
    Load entry points for a specific group.

    See :py:function:`get_entry_points` for further reference on
    additional keyword arguments.

    :param str group_name: the name of the entry point group
    :param bool strict: whether to raise or warn on error
    :returns: mapping from entry point name to loaded entry point
    :rtype: dict
    """
    loaded_entry_points = {}
    for name, entry_point in get_entry_points(
        group_name, strict=strict, **kwargs
    ).items():
        try:
            loaded_entry_points[name] = entry_point.load()
        except Exception as e:  # noqa: F841
            msg = f"Failed to load entry point '{name}': {e}"
            if strict:
                raise RuntimeError(msg)
            logger.warning(msg)
    return loaded_entry_points



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import signal

from rosidl_cli.command.generate import GenerateCommand
from rosidl_cli.command.translate import TranslateCommand
from rosidl_cli.common import get_first_line_doc


def add_subparsers(parser, cli_name, commands):
    """
    Create argparse subparser for each command.

    The ``cli_name`` is used for the title and description of the
    ``add_subparsers`` function call.

    For each command a subparser is created.

    :param parser: the parent argument parser
    :type parser: :py:class:`argparse.ArgumentParser`
    :param str cli_name: name of the command line command to which the
      subparsers are being added
    :param commands: each of the commands contributing specific arguments
    :type commands: :py:class:`List[Command]`
    """
    # add subparser with description of available subparsers
    description = ''

    commands = sorted(commands, key=lambda command: command.name)
    max_length = max(len(command.name) for command in commands)
    for command in commands:
        description += '%s  %s\n' % (
            command.name.ljust(max_length),
            get_first_line_doc(command))
    subparser = parser.add_subparsers(
        title='Commands', description=description,
        metavar=f'Call `{cli_name} <command> -h` for more detailed usage.')
    subparser.dest = '_command'
    subparser.required = True

    # add extension specific sub-sub-parser with its arguments
    for command in commands:
        command_parser = subparser.add_parser(
            command.name,
            description=get_first_line_doc(command),
            formatter_class=argparse.RawDescriptionHelpFormatter)
        command_parser.set_defaults(_command=command)
        command.add_arguments(command_parser)

    return subparser


def main():
    script_name = 'rosidl'
    description = f'{script_name} is an extensible command-line tool ' \
        'for ROS interface generation.'

    # top level parser
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    commands = [GenerateCommand(), TranslateCommand()]

    # add arguments for command extension(s)
    add_subparsers(
        parser,
        script_name,
        commands
    )

    # register argcomplete hook if available
    try:
        from argcomplete import autocomplete
    except ImportError:
        pass
    else:
        autocomplete(parser, exclude=['-h', '--help'])

    # parse the command line arguments
    args = parser.parse_args()

    # call the main method of the command
    try:
        rc = args._command.main(args=args)
    except KeyboardInterrupt:
        rc = signal.SIGINT
    except (ValueError, RuntimeError) as e:
        rc = str(e)
    return rc



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rosidl_cli.extensions import Extension
from rosidl_cli.extensions import load_extensions


class TranslateCommandExtension(Extension):
    """
    The extension point for interface definition translation.

    The following attributes must be defined
    * `input_format`
    * `output_format`

    The following methods must be defined:
    * `translate`
    """

    def translate(
        self,
        package_name,
        interface_files,
        include_paths,
        output_path
    ):
        """
        Translate interface definition files.

        The path to an interface definition file is a relative path optionally
        prefixed by an absolute path followed by a colon ':', in which case
        path resolution is to be performed against that absolute path.

        On output, the directory structure specified by this relative path
        will be replicated e.g. an ``msg/Empty.foo`` file will result in a
        ``msg/Empty.bar`` file under `output_path`.

        :param package_name: name of the package `interface_file` belongs to
        :param interface_files: list of paths to interface definition files
        :param include_paths: list of paths to include dependency interface
          definition files from
        :param output_path: path to directory to hold translated interface
          definition files
        :returns: list of paths to translated interface definition files
        """
        raise NotImplementedError()


def load_translate_extensions(**kwargs):
    """Load extensions for interface definition translation."""
    return load_extensions(
        'rosidl_cli.command.translate.extensions', **kwargs
    )



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from rosidl_cli.command import Command

from .api import generate


class GenerateCommand(Command):
    """Generate source code from interface definition files."""

    name = 'generate'

    def add_arguments(self, parser):
        parser.add_argument(
            '-o', '--output-path', metavar='PATH',
            type=pathlib.Path, default=None,
            help=('Path to directory to hold generated '
                  "source code files. Defaults to '.'."))
        parser.add_argument(
            '-t', '--type', metavar='TYPE',
            dest='types', action='append', default=[],
            help='Target type representations for generation.')
        parser.add_argument(
            '-ts', '--type-support', metavar='TYPESUPPORT',
            dest='typesupports', action='append', default=[],
            help='Target type supports for generation.')
        parser.add_argument(
            '-I', '--include-path', type=pathlib.Path, metavar='PATH',
            dest='include_paths', action='append', default=[],
            help='Paths to include dependency interface definition files from.')
        parser.add_argument(
            'package_name', help='Name of the package to generate code for')
        parser.add_argument(
            'interface_files', metavar='interface_file', nargs='+',
            help=('Relative path to an interface definition file. '
                  "If prefixed by another path followed by a colon ':', "
                  'path resolution is performed against such path.'))

    def main(self, *, args):
        generate(
            package_name=args.package_name,
            interface_files=args.interface_files,
            include_paths=args.include_paths,
            output_path=args.output_path,
            types=args.types,
            typesupports=args.typesupports
        )
